/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim: set ts=8 sts=2 et sw=2 tw=80: */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "RTCCertService.h"
#include "mozilla/net/SocketProcessBridgeChild.h"
#include "mozilla/ipc/BackgroundChild.h"
#include "mozilla/ipc/Endpoint.h"
#include "mozilla/ipc/PBackgroundChild.h"

namespace mozilla::dom {

NS_IMPL_ISUPPORTS(RTCCertService, nsIRTCCertService)

already_AddRefed<nsIRTCCertService> NewRTCCertService() {
  nsCOMPtr<nsIRTCCertService> certService(new RTCCertService());
  certService->Initialize();
  return certService.forget();
}

void RTCCertService::Initialize() {
  using EndpointPromise =
      MozPromise<mozilla::ipc::Endpoint<PRTCCertServiceTransactionChild>,
                 nsCString, true>;
  mInitPromise =
      net::SocketProcessBridgeChild::GetSocketProcessBridge()
          ->Then(
              GetCurrentSerialEventTarget(), __func__,
              [](const RefPtr<net::SocketProcessBridgeChild>& aBridge) {
                mozilla::ipc::Endpoint<PRTCCertServiceTransactionParent>
                    parentEndpoint;
                mozilla::ipc::Endpoint<PRTCCertServiceTransactionChild>
                    childEndpoint;

                mozilla::dom::PRTCCertServiceTransaction::CreateEndpoints(
                    &parentEndpoint, &childEndpoint);

                if (!aBridge || !aBridge->SendInitRTCCertServiceTransaction(
                                    std::move(parentEndpoint))) {
                  NS_WARNING(
                      "RTCCertService async init failed! Webrtc "
                      "networking "
                      "will not work!");
                  return EndpointPromise::CreateAndReject(
                      nsCString("SendInitRTCCertServiceTransaction failed!"),
                      __func__);
                }
                return EndpointPromise::CreateAndResolve(
                    std::move(childEndpoint), __func__);
              },
              [](const nsCString& aError) {
                return EndpointPromise::CreateAndReject(aError, __func__);
              })
          ->Then(
              GetCurrentSerialEventTarget(), __func__,
              [this, self = RefPtr<RTCCertService>(this)](
                  mozilla::ipc::Endpoint<PRTCCertServiceTransactionChild>&&
                      aEndpoint) {
                RefPtr<RTCCertServiceTransactionChild> child =
                    new RTCCertServiceTransactionChild();
                aEndpoint.Bind(child);
                mChild = child;

                return InitPromise::CreateAndResolve(true, __func__);
              },
              [=](const nsCString& aError) {
                NS_WARNING(
                    "RTCCertService async init failed! Webrtc "
                    "networking "
                    "will not work!");
                return InitPromise::CreateAndReject(aError, __func__);
              });
}

RefPtr<RTCCertFingerprintPromise> RTCCertService::GenerateCertificate(
    const nsTArray<uint8_t>& aParam, PRTime aExpires, uint32_t aMechanism,
    uint32_t aSignatureAlg) {
  return mInitPromise->Then(
      GetCurrentSerialEventTarget(), __func__,
      [self = RefPtr<RTCCertService>(this), this, param = aParam.Clone(),
       aExpires, aMechanism, aSignatureAlg](bool /* dummy */) {
        if (!mChild) {
          return RTCCertFingerprintPromise::CreateAndReject(NS_ERROR_FAILURE,
                                                            __func__);
        }
        RefPtr<RTCCertFingerprintPromise> promise =
            mChild
                ->SendGenerateCertificate(param, aExpires, aMechanism,
                                          aSignatureAlg)
                ->Then(
                    GetCurrentSerialEventTarget(), __func__,
                    [](CertFingerprint&& aCertFingerprint) {
                      return RTCCertFingerprintPromise::CreateAndResolve(
                          std::move(aCertFingerprint), __func__);
                    },
                    [](mozilla::ipc::ResponseRejectReason aReason) {
                      return RTCCertFingerprintPromise::CreateAndReject(
                          NS_ERROR_FAILURE, __func__);
                    });
        return promise;
      },
      [](const nsCString& aError) {
        return RTCCertFingerprintPromise::CreateAndReject(NS_ERROR_FAILURE,
                                                          __func__);
      });
}

void RTCCertService::RemoveCertificate(
    const mozilla::dom::CertFingerprint aCertFingerprint) {
  mInitPromise->Then(
      GetCurrentSerialEventTarget(), __func__,
      [self = RefPtr<RTCCertService>(this), this,
       aCertFingerprint](bool /* dummy */) {
        if (mChild) {
          mChild->SendRemoveCertificate(aCertFingerprint);
        }
      },
      [](const nsCString& aError) {});
}

RefPtr<RTCCertificatePromise> RTCCertService::GetCertificate(
    const CertFingerprint aCertFingerprint) {
  return mInitPromise->Then(
      GetCurrentSerialEventTarget(), __func__,
      [self = RefPtr<RTCCertService>(this), this,
       aCertFingerprint](bool /* dummy */) {
        if (!mChild) {
          return RTCCertificatePromise::CreateAndReject(NS_ERROR_FAILURE,
                                                        __func__);
        }
        RefPtr<RTCCertificatePromise> promise =
            mChild->SendGetCertificate(aCertFingerprint)
                ->Then(
                    GetCurrentSerialEventTarget(), __func__,
                    [](const CertDataIPC& aCertDataIPC) {
                      return RTCCertificatePromise::CreateAndResolve(
                          MakeUnique<CertData>(&aCertDataIPC), __func__);
                    },

                    [](mozilla::ipc::ResponseRejectReason aReason) {
                      return RTCCertificatePromise::CreateAndReject(
                          NS_ERROR_FAILURE, __func__);
                    });
        return promise;
      },
      [](const nsCString& aError) {
        return RTCCertificatePromise::CreateAndReject(NS_ERROR_FAILURE,
                                                      __func__);
      });
}

}  // namespace mozilla::dom
