/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef mozilla_dom_RTCCertServiceGlobal_h_
#define mozilla_dom_RTCCertServiceGlobal_h_

#include "ScopedNSSTypes.h"
#include "ipc/IPCMessageUtils.h"
#include "mozilla/ipc/IPDLParamTraits.h"
#include "mozilla/MozPromise.h"

namespace mozilla {
namespace dom {

struct CertFingerprint {
  CertFingerprint() = default;
  explicit CertFingerprint(const nsTArray<uint8_t>& aCertFingerprint);
  operator nsTArray<uint8_t>();

  bool Match(const struct CertFingerprint* aCertFingerprint);
  unsigned char* AsChar() { return reinterpret_cast<unsigned char*>(mHash); }

 public:
  const static size_t sHashByteLen = 16;
  uint64_t mHash[2];
};

struct CertData;
struct CertDataIPC {
  CertDataIPC() = default;
  explicit CertDataIPC(const CertData* aCertData);

 public:
  nsTArray<uint8_t> mCertificate;
  PRTime mExpires;
};

struct CertData {
  CertData(UniqueCERTCertificate aCertificate, PRTime aExpires)
      : mCertificate(std::move(aCertificate)), mExpires(aExpires) {}
  explicit CertData(const CertDataIPC* aCertDataIPC);

  // Don't copy CertData
  CertData(const CertData&) = delete;
  CertData& operator=(const CertData&) = delete;

 public:
  UniqueCERTCertificate mCertificate;
  PRTime mExpires;
};

using RTCCertFingerprintPromise =
    MozPromise<CertFingerprint, nsresult, /* IsExclusive = */ true>;
using RTCCertificatePromise =
    MozPromise<UniquePtr<CertData>, nsresult, /* IsExclusive = */ true>;

void SerializeRSAParam(nsTArray<uint8_t>* aParams,
                       PK11RSAGenParams* aRsaParams);
PK11RSAGenParams DeserializeRSAParam(nsTArray<uint8_t>* aParams);

bool SerializeECParams(nsTArray<uint8_t>* aParams, SECItem* aECParams);
SECItem* DeserializeECParams(nsTArray<uint8_t>* aParams);
}  // namespace dom

namespace ipc {
template <>
struct IPDLParamTraits<dom::CertFingerprint> {
  typedef dom::CertFingerprint paramType;
  static void Write(IPC::MessageWriter* aWriter, ipc::IProtocol* aActor,
                    const paramType& aVar) {
    WriteIPDLParam(aWriter, aActor, aVar.mHash[0]);
    WriteIPDLParam(aWriter, aActor, aVar.mHash[1]);
  }
  static bool Read(IPC::MessageReader* aReader, mozilla::ipc::IProtocol* aActor,
                   paramType* aVar) {
    if (!ReadIPDLParam(aReader, aActor, aVar->mHash) ||
        !ReadIPDLParam(aReader, aActor, aVar->mHash + 1)) {
      return false;
    }
    return true;
  }
};

template <>
struct IPDLParamTraits<dom::CertDataIPC> {
  typedef dom::CertDataIPC paramType;
  static void Write(IPC::MessageWriter* aWriter, ipc::IProtocol* aActor,
                    const paramType& aVar) {
    WriteIPDLParam(aWriter, aActor, aVar.mCertificate);
    WriteIPDLParam(aWriter, aActor, aVar.mExpires);
  }
  static bool Read(IPC::MessageReader* aReader, mozilla::ipc::IProtocol* aActor,
                   paramType* aVar) {
    if (!ReadIPDLParam(aReader, aActor, &aVar->mCertificate) ||
        !ReadIPDLParam(aReader, aActor, &aVar->mExpires)) {
      return false;
    }
    return true;
  }
};
}  // namespace ipc
}  // namespace mozilla

#endif  // mozilla_dom_RTCCertServiceGlobal_h_
